(* Title:      Algebras for Aggregation and Minimisation
   Author:     Walter Guttmann
   Maintainer: Walter Guttmann <walter.guttmann at canterbury.ac.nz>
*)

section {* Algebras for Aggregation and Minimisation *}

theory Aggregation_Algebras

imports "../Stone_Kleene_Relation_Algebras/Kleene_Relation_Algebras"

begin

context sup
begin

no_notation
  sup (infixl "+" 65)

end

context plus
begin

notation
  plus (infixl "+" 65)

end

class sum =
  fixes sum :: "'a \<Rightarrow> 'a"

class s_algebra = stone_relation_algebra + plus + sum +
  assumes sum_plus_right_isotone: "x \<noteq> bot \<and> sum x \<le> sum y \<longrightarrow> z + sum x \<le> z + sum y"
  assumes sum_bot: "sum x + sum bot = sum x"
  assumes sum_plus: "sum x + sum y = sum (x \<squnion> y) + sum (x \<sqinter> y)"
  assumes sum_conv: "sum (x\<^sup>T) = sum x"
begin

lemma sum_disjoint:
  assumes "x \<sqinter> y = bot"
    shows "sum ((x \<squnion> y) \<sqinter> z) = sum (x \<sqinter> z) + sum (y \<sqinter> z)"
  by (subst sum_plus) (metis assms inf.sup_monoid.add_assoc inf.sup_monoid.add_commute inf_bot_left inf_sup_distrib2 sum_bot)

lemma sum_disjoint_3:
  assumes "w \<sqinter> x = bot"
      and "w \<sqinter> y = bot"
      and "x \<sqinter> y = bot"
    shows "sum ((w \<squnion> x \<squnion> y) \<sqinter> z) = sum (w \<sqinter> z) + sum (x \<sqinter> z) + sum (y \<sqinter> z)"
  by (metis assms inf_sup_distrib2 sup_idem sum_disjoint)

lemma sum_symmetric:
  assumes "y = y\<^sup>T"
    shows "sum (x\<^sup>T \<sqinter> y) = sum (x \<sqinter> y)"
  by (metis assms sum_conv conv_dist_inf)

end

class minatom =
  fixes minatom :: "'a \<Rightarrow> 'a"

class m_algebra = s_algebra + minatom +
  assumes minatom_below: "minatom x \<le> --x"
  assumes minatom_atom: "x \<noteq> bot \<longrightarrow> atom (minatom x)"
  assumes minatom_min: "atom y \<and> y \<sqinter> x \<noteq> bot \<longrightarrow> sum (minatom x \<sqinter> x) \<le> sum (y \<sqinter> x)"
begin

lemma minatom_bot:
  "minatom bot = bot"
  by (metis bot_unique minatom_below regular_closed_bot)

lemma minatom_bot_iff:
  "minatom x = bot \<longleftrightarrow> x = bot"
  using covector_bot_closed inf_bot_right minatom_atom vector_bot_closed minatom_bot by fastforce

lemma minatom_meet_bot:
  assumes "minatom x \<sqinter> x = bot"
    shows "minatom x = bot"
proof -
  have "minatom x \<le> -x"
    using assms pseudo_complement by auto
  thus ?thesis
    by (metis minatom_below inf_absorb1 inf_import_p inf_p)
qed

lemma minatom_meet_bot_minatom_iff:
  "minatom x \<sqinter> x = bot \<longleftrightarrow> minatom x = bot"
  using comp_inf.semiring.mult_not_zero minatom_meet_bot by blast

lemma minatom_meet_bot_iff:
  "minatom x \<sqinter> x = bot \<longleftrightarrow> x = bot"
  using inf_bot_right minatom_bot_iff minatom_meet_bot by blast

lemma minatom_regular:
  "regular (minatom x)"
proof (cases "x = bot")
  assume "x = bot"
  thus ?thesis
    by (simp add: minatom_bot)
next
  assume "x \<noteq> bot"
  thus ?thesis
    by (simp add: atom_regular minatom_atom)
qed

lemma minatom_selection:
  "selection (minatom x \<sqinter> y) y"
  using inf_assoc minatom_regular selection_closed_id by auto

(*
lemma sum_bot: "sum bot = bot" nitpick [expect=genuine] oops
lemma plus_bot: "x + bot = x" nitpick [expect=genuine] oops
lemma "sum x = bot \<longrightarrow> x = bot" nitpick [expect=genuine] oops
*)

end

class m_kleene_algebra = m_algebra + stone_kleene_relation_algebra

end

